/*
  composite.c

  Copyright (C) 2015 Taiji Yamada <taiji@aihara.co.jp>
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <math.h>
#include "prime_util.h"

#define T_CMP(T) ((*(T *)a < *(T *)b) ? -1 : (*(T *)a > *(T *)b) ? 1 : 0)
int ull_cmp(const void *a, const void *b) { return T_CMP(unsigned long long int); }
int ull_cmp_r(void *thunk, const void *a, const void *b) { (void)thunk; return T_CMP(unsigned long long int); }

int usage(FILE *fp, int argc, char *argv[])
{
  (void)argc;
  fprintf(fp, "Usage:\n\t%s [-c number ...] [-v number ...] [-w number ...] [-m times]\n", basename(argv[0]));
  return fp == stderr ? 1 : 0;
}
int main(int argc, char *argv[])
{
  huge_uint *constants = NULL, *times = NULL, *composites = NULL;
  size_t n_constants = 0, n_times = 0, n_composites = 0;
  _Bool *zero;
  size_t n_zero = 0;
  int a;
  size_t i, j, k, m = 5;

  for (a=1; a<argc; ++a)
    if (strcmp(argv[a], "-h") == 0)
      return usage(stdout, argc, argv);
    else if (strcmp(argv[a], "-c") == 0 && ++a < argc)
      for (; a<argc; ++a) {
	huge_uint n;
	if ((n = atoll(argv[a]))) {}
	else { --a; break; }
	if ((constants = constants ? realloc(constants, sizeof(*constants)*(n_constants+1)) : malloc(sizeof(*constants)*(n_constants+1))))
	  constants[n_constants++] = n;
      }
    else if (strcmp(argv[a], "-v") == 0 && ++a < argc)
      for (; a<argc; ++a) {
	huge_uint n;
	if ((n = atoll(argv[a]))) {}
	else { --a; break; }
	if ((times = times ? realloc(times, sizeof(*times)*(n_times+1)) : malloc(sizeof(*times)*(n_times+1))))
	  times[n_times++] = n;
	if ((zero = zero ? realloc(zero, sizeof(*zero)*(n_zero+1)) : malloc(sizeof(*zero)*(n_zero+1))))
	  zero[n_zero++] = 0;
      }
    else if (strcmp(argv[a], "-w") && ++a < argc)
      for (; a<argc; ++a) {
	huge_uint n;
	if ((n = atoll(argv[a]))) {}
	else { --a; break; }
	if ((times = times ? realloc(times, sizeof(*times)*(n_times+1)) : malloc(sizeof(*times)*(n_times+1))))
	  times[n_times++] = n;
	if ((zero = zero ? realloc(zero, sizeof(*zero)*(n_zero+1)) : malloc(sizeof(*zero)*(n_zero+1))))
	  zero[n_zero++] = !0;
      }
    else if (strcmp(argv[a], "-m") == 0 && a+1 < argc)
      m = atoll(argv[++a]);
    else
      break;

  huge_uint c = 1;
  for (k=0; k<n_constants; ++k)
    c *= constants[k];

  size_t t = 1;
  for (k=0; k<n_times; ++k)
    t *= m;

  j = 0;
  for (i=0; i<t; ++i) {
    huge_uint n = c;
    for (k=0; k<n_times; ++k) {
      huge_uint p;
      size_t d = (i % (size_t)pow((double)m, (double)(k+1))) /
	(size_t)pow((double)m, (double)k);
      p = power(times[k], (huge_uint)(zero[k] ? d : d+1));
      n *= p;
    }
    if ((composites = composites ? realloc(composites, sizeof(*composites)*(n_composites+1)) : malloc(sizeof(*composites)*(n_composites+1))))
      composites[n_composites++] = n;
  }

  qsort(composites, n_composites, sizeof(huge_uint), ull_cmp);

  for (k=0; k<n_composites; ++k)
    printf("%llu\n", composites[k]);

  return 0;
}
