/*
  prime_polynomial_progression.c

  Copyright (C) 2015 Taiji Yamada <taiji@aihara.co.jp>
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include "prime_util.h"

int usage(FILE *fp, int argc, char *argv[])
{
  (void)argc;
  fprintf(fp, "Usage:\n\t%s [-l lower] [-u upper] [-c coefficient ...] [-k number]\n", basename(argv[0]));
  return fp == stderr ? 1 : 0;
}
int main(int argc, char *argv[])
{
  int a, fd;
  huge_uint n, l = 1, u = 1000000000, k = 17, i, n_return;
  isprime_bit_table_element_t *isprime_bit_table;
  huge_uint **c = NULL;
  size_t n_c = 0, *m_c = NULL;

  for (a=1; a<argc; ++a)
    if (strcmp(argv[a], "-h") == 0)
      return usage(stdout, argc, argv);
    else if (strcmp(argv[a], "-l") == 0 && a+1 < argc)
      l = atoll(argv[++a]);
    else if (strcmp(argv[a], "-u") == 0 && a+1 < argc)
      u = atoll(argv[++a]);
    else if (strcmp(argv[a], "-c") == 0 && ++a < argc) {
      if ((c = c ? realloc(c, sizeof(huge_uint *)*(n_c+1)) : malloc(sizeof(huge_uint *)*(n_c+1))))
	c[n_c] = NULL;
      if ((m_c = m_c ? realloc(m_c, sizeof(size_t)*(n_c+1)) : malloc(sizeof(size_t)*(n_c+1))))
	m_c[n_c] = 0;
      if (c && m_c) {
	for (; a<argc; ++a) {
	  huge_uint n;
	  if ((n = atoll(argv[a]))) {}
	  else { --a; break; }
	  c[n_c] = c[n_c] ? realloc(c[n_c], sizeof(huge_uint)*(m_c[n_c]+1)) : malloc(sizeof(huge_uint)*(m_c[n_c]+1));
	  c[n_c][m_c[n_c]++] = n;
	}
	++n_c;
      }
    }
    else if (strcmp(argv[a], "-k") == 0 && a+1 < argc)
      k = atoll(argv[++a]);
    else
      break;

  isprime_bit_table = isprime_bit_table_mmap(u, &n_return, &fd);
  if (!isprime_bit_table) abort();
  primes_bit_continued_eratosthenes_sieve(n_return, isprime_bit_table, u);
  for (; l<u; l = isprime_bit_table_nextprime(u, isprime_bit_table, l)) {
    for (size_t t=0; t<n_c; ++t) {
      c[t][0] = l;
      for (i=0; ; ++i) {
	n = 0;
	for (size_t j=0; j<m_c[t]; j++) {
	  huge_uint x;
	  x = pow(i, j);
	  n += c[t][j]*x;
	}
	if (!(n < max(n_return, u))) abort();
	if (!isprime_bit_table[n]) break;
      }
      if (i>k) {
	printf("%llu\t", i);
	for (size_t j=0; j<m_c[t]; j++) printf("%llu\t", c[t][j]);
	printf("\n");
      }
    }
  }
  return 0;
}
